{ FIXPATH - (C) 2004-2012 by Bill Stewart (bstewart@iname.com)

  Simple Win32 console utility to fix the Path on Windows 2000 or later.

  The system path is stored in the Path value in the following registry key:

    HKLM\SYSTEM\CurrentControlSet\Control\Session Manager\Environment

  The Path value must be REG_EXPAND_SZ and must contain:

    %SystemRoot%\system32;%SystemRoot%;%SystemRoot%\system32\Wbem

  In the case of Windows 7/2008 R2 or later, we also need:

    %SystemRoot%\system32\WindowsPowerShell\v1.0\

  We have a problem if the Path value:

  1. Is missing.
  2. Is not type REG_EXPAND_SZ.
  3. Doesn't contain the minimum required directories (above).

  In case of (1), we create the Path entry with the minimum required 
  directories.

  In case of (2), we re-write the registry value as REG_EXPAND_SZ.

  In case of (3), we put the minimum required directories at the beginning.

  If user doesn't have sufficient access, Win32 functions will fail.

  A reboot is required for the system to read the corrected Path value.

  Version history:

  1.0.0.0 (2004 March 5):

  *  Initial version.

  2.0.0.0 (2004 August 10):

  *  Minimum required directories can be anywhere in the Path.
  *  Program exits with last Win32 exit code (0 = no error).
  *  Can run in "hands-off" mode by using /Q command-line option.

  3.0.0.0 (2010 November 16):

  * Pauses for an Enter keystroke if /Q not used.
  * Internal corrections: SizeOf(Char) instead of 1; Win32 data types

  4.0.0.0 (2012 August 11):

  * Refactored and simplified code
  * Wrapper functions for registry read and write
  * Added GetOSVersion function
  * For versions >= NT 6.1 (Win 7/Server 2008R2), added PowerShell path
  * Added output of localized error messages (GetErrorMessage function)

}

Program FixPath;

{$APPTYPE CONSOLE}
{$MODE OBJFPC}

Uses
  Strings, Windows;

{$R *.RES}

Const
  // Numeric constants
  NERR_BASE = 2100;
  MAX_NERR = NERR_BASE + 899;
  WINNT_VER50 = $50000;
  WINNT_VER61 = $60001;
  // Typed constants
  SubKey: PChar = 'SYSTEM\CurrentControlSet\Control\Session Manager\Environment';
  ValueName: PChar = 'Path';
  // For OS versions less than 6.1 (Win7/Server 2008 R2)
  MinRequiredDir1: PChar = '%SystemRoot%\system32;%SystemRoot%;%SystemRoot%\system32\Wbem';
  // For OS versions >= 6.1 (Win7/Server 2008 R2 or newer)
  MinRequiredDir2: PChar = '%SystemRoot%\system32;%SystemRoot%;%SystemRoot%\system32\Wbem;%SystemRoot%\system32\WindowsPowerShell\v1.0\';

Var
  QuietMode: Boolean;           // /Q parameter used?
  OSVersion: DWORD;             // current OS version
  pMinimumDirs: PChar;          // minimum required directories
  RegistryUpdated: Boolean;     // was registry updated?
  GetResult, SetResult: DWORD;  // Get/Set registry data result code
  pCurrValueData: PChar;        // current value of registry data
  ValueType: DWORD;             // registry value type
  CurrValueLength: DWORD;       // length of current value
  pNewValueData: PChar;         // new value
  NewValueLength: DWORD;        // length of new value
  ValueTypeOK: Boolean;         // is value type REG_EXPAND_SZ?
  MinRequiredFound: Boolean;    // minimum required data found?

///////////////////////////////////////////////////////////////////////////////
// Returns a DWORD as a String
///////////////////////////////////////////////////////////////////////////////
Function DWORDToString(Const D: DWORD): String;
  Begin
  Str(D, Result);
  End;

///////////////////////////////////////////////////////////////////////////////
// Returns the localized string interpretation of the specified error code
///////////////////////////////////////////////////////////////////////////////
Function GetErrorMessage(ErrorNumber: DWORD): String;
  Var
    FormatFlags: DWORD;
    ModuleHandle: HMODULE;
    pBuffer: PChar;
  Begin
  FormatFlags := FORMAT_MESSAGE_ALLOCATE_BUFFER Or FORMAT_MESSAGE_IGNORE_INSERTS Or FORMAT_MESSAGE_FROM_SYSTEM;
  ModuleHandle := 0;
  If (ErrorNumber >= NERR_BASE) And (ErrorNumber <= MAX_NERR) Then
    Begin
    ModuleHandle := LoadLibraryEx('netmsg.dll', 0, LOAD_LIBRARY_AS_DATAFILE);
    If ModuleHandle <> 0 Then
      FormatFlags := FormatFlags Or FORMAT_MESSAGE_FROM_HMODULE;
    End;
  If FormatMessage(FormatFlags,
                   Pointer(ModuleHandle),
                   ErrorNumber,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   @pBuffer,
                   0,
                   Nil) > 0 Then
    Begin
    Result := StrPas(pBuffer);
    LocalFree(HLOCAL(pBuffer));
    End
  Else
    Result := 'Error ' + DWORDToString(ErrorNumber);
  If ModuleHandle <> 0 Then
    FreeLibrary(ModuleHandle);
  End;

///////////////////////////////////////////////////////////////////////////////
// Prints informational message
///////////////////////////////////////////////////////////////////////////////
Procedure InfoMessage(pMinimumDirs: PChar);
  Begin
  WriteLn('FIXPATH 4.0 - (C) 2004-2012 by Bill Stewart (bstewart@iname.com)');
  WriteLn;
  WriteLn('The system path is stored in the Path value in the following registry key:');
  WriteLn;
  WriteLn('  HKLM\System\CurrentControlSet\Control\Session Manager\Environment');
  WriteLn;
  WriteLn('The Path registry value must be of type REG_EXPAND_SZ and must contain, at a');
  WriteLn('minimum, the following directories:');
  WriteLn;
  WriteLn(pMinimumDirs);
  WriteLn;
  WriteLn('If either of these conditions are false, then the Path is not correct. This');
  WriteLn('program attempts to correct both of these problems, if found.');
  WriteLn;
  WriteLn('You must be logged on as a member of Administrators to be able to update the');
  WriteLn('Path. On Windows Vista or later with UAC enabled, you must run this program');
  WriteLn('elevated (right-click, Run as administrator).');
  WriteLn;
  WriteLn('You may run this program without user interaction by specifying /Q on the');
  WriteLn('command line.');
  End;

///////////////////////////////////////////////////////////////////////////////
// Returns the current OS version, or 0 for "fail"
///////////////////////////////////////////////////////////////////////////////
Function GetOSVersion: DWORD;
  Var
    OSV: OSVERSIONINFO;
    OK: BOOL;
  Begin
  Result := 0;
  OSV.dwOSVersionInfoSize := SizeOf(OSVERSIONINFO);
  OK := GetVersionEx(OSV);
  If OK Then
    Begin
    If OSV.dwPlatformId = VER_PLATFORM_WIN32_NT Then
      Result := (OSV.dwMajorVersion Shl $10) Or OSV.dwMinorVersion;
    End;
  End;

///////////////////////////////////////////////////////////////////////////////
// Wrapper for RegQueryValueEx
///////////////////////////////////////////////////////////////////////////////
Function RegGetValue2(hKey: HKEY;
                      pSubKey,
                      pValueName: PChar;
                      Var ValueType: DWORD;
                      Var pData: Pointer;
                      Var cbData: DWORD): LONG;
  Var
    hkHandle: DWORD;
  Begin
  Result := RegOpenKeyEx(hKey,
                         pSubKey,
                         0,
                         KEY_READ,
                         hkHandle);
  If Result = ERROR_SUCCESS Then
    Begin
    // First call with lpType and lpData NULL: retrieve required size into lpcbData
    Result := RegQueryValueEx(hkHandle,
                              pValueName,
                              Nil,
                              Nil,
                              Nil,
                              @cbData);
    If Result = ERROR_SUCCESS Then
      Begin
      // Now, allocate the buffer and retrieve the data into it
      GetMem(pData, cbData);
      Result := RegQueryValueEx(hkHandle,
                                pValueName,
                                Nil,
                                @ValueType,
                                pData,
                                @cbData);
      If Result <> ERROR_SUCCESS Then
        Begin
        ValueType := 0;
        If (pData <> Nil) And (cbData > 0) Then
          FreeMem(pData, cbData);
        pData := Nil;
        cbData := 0;
        End;
      End;
    RegCloseKey(hkHandle);
    End;
  End;

///////////////////////////////////////////////////////////////////////////////
// Wrapper for RegSetValueEx
///////////////////////////////////////////////////////////////////////////////
Function RegSetValue2(hKey: HKEY;
                      pSubKey,
                      pValueName: PChar;
                      ValueType: DWORD;
                      pData: Pointer;
                      cbData: DWORD): LONG;
  Var
    hkHandle: DWORD;
  Begin
  Result := RegOpenKeyEx(hKey,
                         pSubKey,
                         0,
                         KEY_ALL_ACCESS,
                         hkHandle);
  If Result = ERROR_SUCCESS Then
    Result := RegSetValueEx(hkHandle,
                            pValueName,
                            0,
                            ValueType,
                            pData,
                            cbData);
  RegCloseKey(hkHandle);
  End;

///////////////////////////////////////////////////////////////////////////////
// Pauses for an Enter keystroke with a prompt
///////////////////////////////////////////////////////////////////////////////
Procedure Pause;
  Begin
  WriteLn;
  Write('Press Enter to continue . . . ');
  ReadLn;
  End;

///////////////////////////////////////////////////////////////////////////////
// Writes the specified prompt, return True if the user entered 'Y'
///////////////////////////////////////////////////////////////////////////////
Function Query(Const Prompt: String): Boolean;
  Var
    Ch: Char;
  Begin
  Repeat
    Write(Prompt, ' (Y/N) ');
    ReadLn(Ch);
    Ch := UpCase(Ch);
  Until (Ch = 'Y') Or (Ch = 'N');
  Result := Ch = 'Y';
  End;

///////////////////////////////////////////////////////////////////////////////
// Main
///////////////////////////////////////////////////////////////////////////////
Begin
  QuietMode := UpCase(ParamStr(1)) = '/Q';

  OSVersion := GetOSVersion;

  If OSVersion < WINNT_VER61 Then
    pMinimumDirs := MinRequiredDir1
  Else
    pMinimumDirs := MinRequiredDir2;

  If ParamStr(1) = '/?' Then
    Begin
    InfoMessage(pMinimumDirs);
    Pause;
    Exit;
    End;

  If OSVersion < WINNT_VER50 Then
    Begin
    ExitCode := ERROR_OLD_WIN_VERSION;
    WriteLn(GetErrorMessage(ExitCode));
    If Not QuietMode Then
      Pause;
    Exit;
    End;

  If UpCase(ParamStr(1)) <> '/Q' Then
    Begin
    InfoMessage(pMinimumDirs);
    WriteLn;
    If Not Query('Check for errors and attempt to repair now?') Then
      Exit;
    WriteLn;
    End;

  RegistryUpdated := False;

  // Possible GetResult values:
  // ERROR_FILE_NOT_FOUND - registry value is missing
  // ERROR_SUCCESS - retrieved registry value successfully
  // anything else - this is bad; do nothing

  GetResult := RegGetValue2(HKEY_LOCAL_MACHINE,
                            SubKey,
                            ValueName,
                            ValueType,
                            pCurrValueData,
                            CurrValueLength);

  If GetResult = ERROR_FILE_NOT_FOUND Then
    Begin
    WriteLn('Error! Path value not found in registry!');
    // SizeOf(Char) - allow for terminating null
    SetResult := RegSetValue2(HKEY_LOCAL_MACHINE,
                              SubKey,
                              ValueName,
                              REG_EXPAND_SZ,
                              pMinimumDirs,
                              StrLen(pMinimumDirs) + SizeOf(Char));
    RegistryUpdated := SetResult = ERROR_SUCCESS;
    If RegistryUpdated Then
      WriteLn('Configured Path value with minimum required directories.');
    End
  Else If GetResult <> ERROR_SUCCESS Then
    Begin
    WriteLn('Error ', GetResult, ' reading from registry');
    WriteLn(GetErrorMessage(GetResult));
    End;

  // Truth table for ValueTypeOK and MinRequiredFound:
  // ValueTypeOK  MinRequiredFound  Action
  // True         True              Nothing
  // False        True              Rewrite pCurrValueData as REG_EXPAND_SZ
  // True         False             Write pMinimumDirs + pCurrValueData
  // False        False             (same as above)

  If GetResult = ERROR_SUCCESS Then
    Begin
    ValueTypeOK := ValueType = REG_EXPAND_SZ;
    If ValueTypeOK Then
      WriteLn('The registry value type is correct (REG_EXPAND_SZ).')
    Else
      WriteLn('Error! The registry value type is not REG_EXPAND_SZ!');
    MinRequiredFound := StrIPos(pCurrValueData, pMinimumDirs) <> Nil;
    If MinRequiredFound Then
      WriteLn('The minimum required directories were found in the Path.')
    Else
      WriteLn('Error! The minimum required directories were not found in the Path!');
    If (Not ValueTypeOK) Or (Not MinRequiredFound) Then
      Begin
      If MinRequiredFound Then
        Begin
        // SizeOf(Char) - allow for terminating null
        NewValueLength := StrLen(pCurrValueData) + SizeOf(Char);
        GetMem(pNewValueData, NewValueLength);
        StrCopy(pNewValueData, pCurrValueData);
        End
      Else
        Begin
        // SizeOf(Char) * 2 - allow for semicolon delimiter and terminating null
        NewValueLength := StrLen(pMinimumDirs) + StrLen(pCurrValueData) + SizeOf(Char) * 2;
        GetMem(pNewValueData, NewValueLength);
        StrCopy(pNewValueData, pMinimumDirs);
        // If current value wasn't blank, append it to minimum required dirs
        If StrComp(pCurrValueData, '') <> 0 Then
          Begin
          // Append semicolon delimiter if needed
          If pCurrValueData[0] <> ';' Then
            StrCat(pNewValueData, ';');
          StrCat(pNewValueData, pCurrValueData);
          End;
        End;
      SetResult := RegSetValue2(HKEY_LOCAL_MACHINE,
                                SubKey,
                                ValueName,
                                REG_EXPAND_SZ,
                                pNewValueData,
                                NewValueLength);
      RegistryUpdated := SetResult = ERROR_SUCCESS;
      FreeMem(pNewValueData, NewValueLength);
      If RegistryUpdated Then
        WriteLn('The Path has been updated successfully.');
      End;
    End;

  If CurrValueLength > 0 Then
    FreeMem(pCurrValueData, CurrValueLength);

  If SetResult <> ERROR_SUCCESS Then
    Begin
    WriteLn('Error ', SetResult, ' setting value in registry');
    WriteLn(GetErrorMessage(SetResult));
    End;

  If RegistryUpdated Then
    Begin
    WriteLn;
    WriteLn('You must reboot for the change to take effect.');
    End;

  If Not QuietMode Then
    Pause;
End.
